# -*- coding: utf-8 -*-
import inspect

from .DomainUtils import DomainUtils


class PatternExecBlockParser:
    def __init__(self, debug=False):
        self.debug = debug

        # dict key   is the pattern exec block name
        # dict value is a list with pattern burst blocks
        self.patt_exec_block2patt_burst = {}

        self.curr_patt_exec = DomainUtils.global_domain

        self.curr_timing_domain = DomainUtils.global_domain

        # List with names of all pattern block which should be executed
        self.used_patterns = []

        # dict key   is the pattern exec block name
        # dict value is the timing domain
        self.patt_exec_block2time_domain = {}

        self.is_patternexec_block_defined = False

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_pattern_exec__PATTERN_EXEC_BLOCK_NAME(self, t):

        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if t in self.patt_exec_block2patt_burst:
            err_msg = f"Pattern exec block {t} already defined! Pattern exec block name must be unique."
            raise Exception(err_msg)
        else:
            self.patt_exec_block2patt_burst[t.value] = []
            self.curr_patt_exec = t.value

    def b_pattern_exec__OPEN_PATTERN_EXEC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any PatternExec block"
            raise Exception(err_msg)
        if self.is_signalgroups_block_defined == False:
            err_msg = "Expected SignalGroups block before any PatternExec block"
            raise Exception(err_msg)
        if self.is_timing_block_defined == False:
            err_msg = "Expected Timing block before any PatternExec block"
            raise Exception(err_msg)
        if self.is_patternburst_block_defined == False:
            err_msg = "Expected PatternBurst block before any PatternExec block"
            raise Exception(err_msg)
        self.is_patternexec_block_defined = True

        global_domain = DomainUtils.global_domain

        if global_domain in self.patt_exec_block2patt_burst:
            err_msg = "Only one global PatternExec block can exists!"
            raise Exception(err_msg)

        if self.curr_patt_exec == global_domain:
            self.patt_exec_block2patt_burst[self.curr_patt_exec] = []

    def b_pattern_exec__PATTERN_BURST_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        patt_burst_name = t.value

        if patt_burst_name not in self.patt_burst_block2pattern_blocks:
            err_msg = f"Pattern burst block named {patt_burst_name} is not defined!"
            raise Exception(err_msg)

        burst_names = self.patt_exec_block2patt_burst[self.curr_patt_exec]
        burst_names.append(patt_burst_name)

        patterns = self.patt_burst_block2pattern_blocks.get(patt_burst_name)
        for pattern in patterns:

            if pattern not in self.used_patterns:
                self.used_patterns.append(pattern)

            if pattern not in self.patt2timing_domain:
                self.patt2timing_domain[pattern] = [self.curr_timing_domain]
            else:
                domains = self.patt2timing_domain[pattern]
                if self.curr_timing_domain not in domains:
                    if domains is None:
                        self.patt2timing_domain[pattern] = [self.curr_timing_domain]
                    else:
                        domains.append(self.curr_timing_domain)

        macro_domain = self.patt_burst2macro_domain[patt_burst_name]
        if macro_domain in self.macro_domain2timing_domain:
            l = self.macro_domain2timing_domain[macro_domain]
            if self.curr_timing_domain not in l:
                l.append(self.curr_timing_domain)
        else:
            self.macro_domain2timing_domain[macro_domain] = [self.curr_timing_domain]

        proc_domain = self.patt_burst2proc_domain[patt_burst_name]
        if proc_domain in self.proc_domain2timing_domain:
            l = self.proc_domain2timing_domain[proc_domain]
            if self.curr_timing_domain not in l:
                l.append(self.curr_timing_domain)
        else:
            self.proc_domain2timing_domain[proc_domain] = [self.curr_timing_domain]

    def b_pattern_exec__TIMING_DOMAIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_timing_domain = t.value

        if self.curr_timing_domain not in self.time_domain2wft:
            err_msg = f"Timing domain {self.curr_timing_domain} is not defined!"
            raise Exception(err_msg)

    def b_pattern_exec__CLOSE_PATTERN_EXEC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.patt_exec_block2time_domain[self.curr_patt_exec] = self.curr_timing_domain

        self.curr_patt_exec = DomainUtils.global_domain
        self.curr_timing_domain = DomainUtils.global_domain



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os

from lark import Lark
from lark.exceptions import UnexpectedToken
from lark.exceptions import UnexpectedCharacters
from lark.exceptions import VisitError

from .STILLark import STILLark

from . import SyntaxParserExceptions
from . import STILSemanticException
from . import utils

class STILParser(STILLark):
    def __init__(self, stil_file, propagate_positions=True, expanding_procs = False, debug=False):

        STILLark.__init__(self, stil_file, debug)

        self.err_line = -1
        self.err_col = -1
        self.err_msg = ""

        self.tree = None

        self.is_parsing_done = False
        # Depends on ATE pattern engine capabilities, procedures can be expanded or not.
        # This choice impacts calculation of the total VA
        self.expanding_procs = expanding_procs

        grammar_base_file = os.path.dirname(__file__)
        # Adding main lark grammar file
        grammar_file = os.path.join(str(grammar_base_file), "grammars", "stil.lark")
        grammars_file = os.path.join(str(grammar_base_file), "grammars")
        # Adding all lark grammars files under grammars folder
        ip = list()
        ip.append(grammars_file)
        with open(grammar_file) as grammar:
            self.parser = Lark(
                grammar.read(),
                start="start",
                parser="lalr",
                propagate_positions=propagate_positions,
                import_paths=ip,
            )
        
        self.is_compressed = utils.check_for_compression(stil_file)
    
    def parse_syntax(self, debug=False):
        """
        Parameters
        ----------
        stil : str
            stil argument can be either location of STIL file or
            the STIL file content as string.
        debug : TYPE, optional
            Debug flag. The default is False.

        Returns
        -------
        tree : Lark tree object

        """
        if debug:
            print("\n===========================")
            print("Start of syntax parsing :\n")

        try:
            if os.path.exists(self.stil_file):
                if self.is_compressed:
                    data = utils.get_uncompressed_data(self.stil_file)
                    self.tree = self.parser.parse(data)
                    if debug == True:
                        print(self.tree.pretty())
                else:
                    with open(self.stil_file) as data:
                        self.tree = self.parser.parse(data.read())
                        if debug == True:
                            print(self.tree.pretty())

            if debug:
                print("\nEnd of syntax parsing")
                print("===========================\n")

            return self.tree

        except UnexpectedToken as e:

#            raise Exception()
#            print(f"UnexpectedToken Exception:\n {e.token.type}")

            if debug:
                print(f"UnexpectedToken Exception:\n {e}")

            line = e.line
            self.err_line = line

            col = e.column
            self.err_col = col
            
            line_with_error = utils.get_line(self.stil_file, e.line)

            self.err_msg = f"\nERROR:\tIn the line number {line}, column {col}:\n"
            self.err_msg += f"\t{line_with_error}\n"
            self.err_msg += utils.get_col_error_pos(col) + "\n"

            if str(e).startswith("Unexpected token Token('$END'"):
                s = str(e).split("\n")
                i = 0
                for s1 in s:
                    i += 1
                    if i == 3:
                        strim = s1.strip()
                        if strim == "*":
                            lines = line_with_error.strip()
                            exp_token = "';' or '}'"
                            if (lines.endswith(";")):
                                exp_token = "'}'"
                            self.err_msg += f"\tEnd of the file reached, but expected {exp_token}"
                        else:
                            if strim.startswith("* "):
                                self.err_msg += "\tUnexpected end of the file or block is not closed (including nested ones) \n"
#                                expected = strim[2:]
#                                self.err_msg += f"\t{SyntaxParserExceptions(debug).transform([expected])}\n"
                            else:
                                self.err_msg += s1 + "\n"
            else:
                self.err_msg += f"\tUnexpected token : {e.token}\n"
                token = e.token.type
                if token == "b_pattern_burst__CLOSE_PATT_OR_BURST":
                    self.err_msg += "\tDid you miss to write the mandatory PatList block ? \n"
                else:
                    self.err_msg += f"\t{SyntaxParserExceptions(debug).transform(e.expected)}\n"

            print(self.err_msg)

        except UnexpectedCharacters as e:

            if debug:
                print(f"UnexpectedCharacters Exception:\n {e}")

            line = e.line
            self.err_line = line

            col = e.column
            self.err_col = col

            line_with_error = utils.get_line(self.stil_file, line)

            self.err_msg = f"\nERROR:\tIn the line number {line}, column {col}:\n"
            self.err_msg += f"\t{line_with_error}\n"
            self.err_msg += utils.get_col_error_pos(col) + "\n"
            self.err_msg += f"\tUnexpected character : {e.char}\n"
            self.err_msg += f"\t{SyntaxParserExceptions(debug).transform(e.allowed)}\n"

            print(self.err_msg)

        return None

    def parse_semantic(
        self,
        # stil,
        debug=False,
    ):

        # if tree is None, syntax parsing is already failed
        if self.tree != None:
            if debug:
                print("\n===========================")
                print("Start of semantic parsing :\n")
            try:
                self.transform(self.tree)
                self.eof()
                self.is_parsing_done = True

                if debug:
                    print("\nEnd of semantic parsing")
                    print("===========================\n")
                                    
            except VisitError as e:

                #raise Exception()
                #if debug:
                #    raise Exception()
                
                line = e.obj.line
                self.err_line = line
                col = e.obj.column
                self.err_col = col
                err_msg = e.orig_exc
                
                if isinstance(e.orig_exc, STILSemanticException):
                    line = e.orig_exc.line
                    self.err_line = line
                    col = e.orig_exc.col
                    self.err_col = col
                    err_msg = e.orig_exc.msg

                line_with_error = utils.get_line(self.stil_file, line)
                self.err_msg = f"\nERROR:\tIn the line number {line}, column {col} [{line},{col}]:\n"
                self.err_msg += f"\t{line_with_error}\n"
                self.err_msg += utils.get_col_error_pos(col) + "\n"
                self.err_msg += f"\t{err_msg}\n"

                print(self.err_msg)

        else:
            print("ERROR : syntax parse must be performed before semantic parsing b")
